//=============================================================================
// Nephthys ServerQuery.
//
// This is the original Kerberos ServerQuery code.
//
// This is some merged stuff based on:
// - original UdpServerQuery
// - my old trials checking out what this all is for ;-P
// - TCP Wolf's UdpServerQueryWFIX (thanx :-))
// - UCop2 UdpServerQueryFIX (thanx :-))
// - my own ideas, especially auto-ban code and using server's mod's ban list
//
// (C) 2004 by Zora
//=============================================================================
class NptServerQuery expands UdpServerQuery
	config(Nephthys);


struct NptAddServerRule
{
	var() config string Tag;
	var() config string Value;
};

var(Report) config NptAddServerRule AddServerRule[20];	// additional customized server rules
var(Report) config	bool	bReportPlayers;
var(Report) config	bool	bReportSpectators;
var(Report) config	bool	bReportBots;
var(Report) config	bool	bReportCameras;
var(Report) config	bool	bReportOthers;
var(Report) config	bool	bCountRealPlayersOnly;		// for server list: don't call specs, etc.

// RQA attacks?!
var(Settings) config	bool	bLogAllRequests;		// debug...
var(Settings) config	bool	bLogRejected;			// log attacker requests
var(Settings) config	bool	bLogInvalidRequests;		// log mal-formed requests
var(Settings) config	bool	bLogInvalidSends;		// log mal-formed replies
var(Settings) config	bool	bLogSendText;			// log all we're sending
var(Settings) config	bool	bLogSendFail;			// log failing sends

var(Settings) config	int	MaxRequestsForRQA;		// if this number of requests from IP within TimeDeltaForRQA then server is under RQA attack
var(Settings) config	float	TimeDeltaForRQA;
var(Settings) config	bool	bBanRQA;			// ban RQA attackers permanently

var(Settings) config	int	MaxRequestsForDRQA;		// if this number of requests within TimeDeltaForDRQA then server is under DRQA attack
var(Settings) config	float	TimeDeltaForDRQA;

// runtime
struct NptAttack {
	var	bool	bActive;
	var	int	Count;			// how many repetitions yet ...
	var	float	DragTime;		// ... since this time
	var	float	FirstTime;		// time stamp of first attack detection
};

var	NptAttack	DRQA;
var NptAttack	RQA;
var string		RqaIP;
var NptAttack	Current;
var string		CurrentIP;

function PostBeginPlay()
{
	Super.PostBeginPlay();

	RQA.bActive = false;
	DRQA.bActive = false;
}

//=============================================================================

function MyLog(string Msg)
{
	Log(Now( Level ) $": "$ Msg, 'Nephthys');
}

const Digits = "0123456789ABCDEF";

function static string IntToHex( int in, optional int format )
{
	local string result;

	do {
		result = Mid( Digits, in & 15, 1 ) $ result;
		in /= 16;
	}
	until ( --format <= 0 && in == 0 );
	return result;
}

function DumpToLog( int size, string str )
{
	local int idx, jdx, lgt;
	local string s;

	for ( idx = 0; idx < size; idx += 16 ) {
		s = IntToHex( idx, 4 ) $ ": ";
		for ( jdx = 0; jdx < 16; jdx++ ) {
			if ( idx + jdx >= size ) s = s $ "   ";
			else s = s $ IntToHex( asc( mid( str, idx + jdx, 1 ) ), 2 ) $ " ";
		}
		for ( jdx = 0; jdx < 16; jdx++ ) {
			if ( idx + jdx >= size ) break;
			else s = s $ mid( str, idx + jdx, 1 );
		}
		MyLog( s );
	}
}

function string DumpToStr( int lgt, string str )
{
	local int idx;
	local string ret, s;

	ret = "";
	s = str;
	for ( idx = 0; idx < lgt; idx++ ) {
		ret = ret $ string(asc( left( s, 1 ) )) $ "  ";
		s = mid( s, 1 );
	}

	ret = ret $ str;
	return ret;
}

// help function to get ip
function string IpOnly(IpAddr addr)
{
	local string ip;

	ip = IpAddrToString( addr );
	return Left( ip, InStr( ip, ":" ) );
}

// help function from MM
function bool isAPlayer( Pawn P )
{
	// this sucks because some game-modes have their own abstract subclass of
	// pawn ignoring the default spectator (e.g. UCop)

	if ( P == None )	return False;	// pff
	if ( !P.bIsPlayer )	return False;
	if ( P.IsA('Camera') || P.IsA('Spectator') ) return False;  // most important
	if ( P.IsA('Bots') )	return True;	// 2nd...
	if ( P.IsA('ScriptedPawn') || P.IsA('FlockMasterPawn') || P.IsA('FlockPawn') ) return False;

	return True;
}

function static string Now( LevelInfo Level )
{
	return Level.Year $ "/" $ Right( "0" $ Level.Month, 2 )
			  $ "/" $ Right( "0" $ Level.Day, 2 )
			  $ " " $ Right( "0" $ Level.Hour, 2 )
			  $ ":" $ Right( "0" $ Level.Minute, 2 )
			  $ ":" $ Right( "0" $ Level.Second, 2 );
}

//=============================================================================

// Return a string of information on a player.
function string GetPlayer2( Pawn P, int PlayerNum, string type )
{
	local int ping;
	local string ResultSet;

	// Name
	ResultSet = "\\player_" $ PlayerNum $ "\\" $ P.PlayerReplicationInfo.PlayerName;

	// Frags
	ResultSet = ResultSet $ "\\frags_" $ PlayerNum $ "\\" $ int(P.PlayerReplicationInfo.Score);

	// Ping
	if (type ~= "Bot")		ResultSet = ResultSet $ "\\ping_" $ PlayerNum $ "\\ BOT";
	else if (type ~= "Camera")	ResultSet = ResultSet $ "\\ping_" $ PlayerNum $ "\\ CAM";
	else {
		if ( P.PlayerReplicationInfo != None )	ping = P.PlayerReplicationInfo.ping;
		else					ping = 9999;
		ResultSet = ResultSet $ "\\ping_" $ PlayerNum $ "\\ " $ ping;
	}

	// Team
	ResultSet = ResultSet $ "\\team_" $ PlayerNum $ "\\" $ P.PlayerReplicationInfo.Team;

	// Skin
	if ( type ~= "Spectator" || type ~= "DemoRecorder" )	ResultSet = ResultSet $ "\\skin_" $ PlayerNum $ "\\" $ type;
	else							ResultSet = ResultSet $ "\\skin_" $ PlayerNum $ "\\" $ GetItemName(string(P.Skin));

	// Mesh
	if ( type ~= "Spectator" || type ~= "DemoRecorder" )	ResultSet = ResultSet $ "\\mesh_" $ PlayerNum $ "\\" $ type;
	else							ResultSet = ResultSet $ "\\mesh_" $ PlayerNum $ "\\" $ GetItemName(string(P.Mesh));

	// Time Playing ?

	return ResultSet;
}

// Send data for each player
function bool SendPlayers( IpAddr Addr, int QueryNum, out int PacketNum, string FinalPacket )
{
	local Pawn P;
	local int idx, inlist, inlist2;
	local bool bResult, bSendResult;
	local Pawn PList[120], PList2[120];
	local string PType[120], PType2[120];
	local PlayerReplicationInfo PRI;

	// MyLog( "SendPlayers called..." );

	// build up the pawnlist to send
	inlist = 0;
	foreach AllActors( class'PlayerReplicationInfo', PRI ) {
		if (PRI.Owner.IsA('Bots')) {
			if ( bReportBots ) {
				PList[inlist] = Pawn(PRI.Owner);
				PType[inlist++] = "Bot";
			}
		}
		else if ( isAPlayer( Pawn(PRI.Owner) ) ) {
			if ( bReportPlayers ) {
				PList[inlist] = Pawn(PRI.Owner);
				PType[inlist++] = "Player";
			}
		}
		else if ( PRI.Owner.IsA('Spectator') ) {
			if ( bReportSpectators ) {
				PList[inlist] = Pawn(PRI.Owner);
				if ( PRI.Owner.IsA('DemoRecSpectator') )	PType[inlist++] = "DemoRecorder";
				else						PType[inlist++] = "Spectator";
			}
		}
		else if ( PRI.Owner.IsA('Camera') ) {
			if ( bReportCameras ) {
				PList[inlist] = Pawn(PRI.Owner);
				PType[inlist++] = "Camera";
			}
		}
		else if ( bReportOthers ) {
			PList[inlist] = Pawn(PRI.Owner);
			PType[inlist++] = "Other";
		}
	}

	// list built

	// sorting list to list 2
	inlist2 = 0;

	// get players
	for ( idx = 0; idx < inlist; idx++ ) {
		if ( PType[idx] == "Player" ) {
			PType2[inlist2] = PType[idx];
			PList2[inlist2++] = PList[idx];
		}
	}

	// get rest
	for ( idx = 0; idx < inlist; idx++ ) {
		if ( PType[idx] != "Player" ) {
			PType2[inlist2] = PType[idx];
			PList2[inlist2++] = PList[idx];
		}
	}

	// list sorted, players now all have low ID numbers, all other "things" higher

	// sending stuff
	bResult = False;

	for ( idx = 0; idx < inlist2; idx++ ) {
		if ( idx == inlist2-1 && FinalPacket == "final" ) {
			bSendResult = SendQueryPacket( Addr, GetPlayer2(PList2[idx], idx, PType2[idx]), QueryNum, ++PacketNum, "final" );
		}
		else {
			bSendResult = SendQueryPacket( Addr, GetPlayer2(PList2[idx], idx, PType2[idx]), QueryNum, ++PacketNum, "" );
		}
		bResult = bSendResult || bResult;
	}

	return bResult;
}

function int GetNumPlayers()
{
	local int NumPlayers;
	local PlayerReplicationInfo PRI;

	foreach AllActors( class'PlayerReplicationInfo', PRI ) {
		if ( PRI.Owner.IsA('Bots') ) {
			if ( bReportBots )	NumPlayers++;
		}
		else if ( isAPlayer( Pawn(PRI.Owner) ) ) {
			if ( bReportPlayers )	NumPlayers++;
		}
		else if ( PRI.Owner.IsA('Spectator') ) {
			if ( bReportSpectators ) NumPlayers++;
		}
		else if ( PRI.Owner.IsA('Camera') ) {
			if ( bReportCameras )	NumPlayers++;
		}
		else if ( bReportOthers )	NumPlayers++;
	}
	return NumPlayers;
}

// Return a string of important system information.
function string GetInfo2( int NumPlayers )
{
	local int idx;
	local string ResultSet;

	// The server name, i.e.: Bob?s Server
	ResultSet = "\\hostname\\"$Level.Game.GameReplicationInfo.ServerName;

	// The short server name
	ResultSet = ResultSet$"\\shortname\\"$Level.Game.GameReplicationInfo.ShortName;

	// The server port.
	ResultSet = ResultSet$"\\hostport\\"$Level.Game.GetServerPort();

	// (optional) The server IP
	// if (ServerIP != "")
	// ResultSet = ResultSet$"\\hostip\\"$ServerIP;

	// The map/level name
	if ( Level.Title == "" || Level.Title ~= "Untitled" )	ResultSet = ResultSet$"\\mapname\\"$getUrlMap();
	else							ResultSet = ResultSet$"\\mapname\\"$Level.Title;

	// The mod or game type
	ResultSet = ResultSet$"\\gametype\\"$Level.Game.GameName;

	// The number of players
	ResultSet = ResultSet$"\\numplayers\\"$NumPlayers;

	// The maximum number of players
	ResultSet = ResultSet$"\\maxplayers\\"$Level.Game.MaxPlayers;

	// The game mode: openplaying
	ResultSet = ResultSet$"\\gamemode\\openplaying";

	// The version of this game.
	ResultSet = ResultSet$"\\gamever\\"$Level.EngineVersion;

	// The most recent network compatible version.
	ResultSet = ResultSet$"\\mingamever\\"$Level.MinNetVersion;

	// add some more info...
	for ( idx = 0; idx < ArrayCount(AddServerRule) && AddServerRule[idx].Tag != ""; idx++ ) {
		ResultSet = ResultSet $ "\\" $ AddServerRule[idx].Tag $ "\\" $ AddServerRule[idx].Value;
	}

	return ResultSet;
}

//=============================================================================

function bool SendText( IpAddr Addr, coerce string Str )
{
	local int lgt;
	local bool bSuccess;

	if ( bLogInvalidSends ) {
		lgt = len( Str );
		if ( lgt >= 512 )  MyLog( "SendText: reply too long (" $ lgt $ ") for in-game serverbrowser (max. 511)" );
	}
	if ( bLogSendText ) {
		lgt = len( Str );
		MyLog( "SendText(" $ IpAddrToString( Addr ) $ ", len=" $ lgt $ " )" );
		DumpToLog( lgt, str );
	}
	bSuccess = Super.SendText( Addr, Str );
	if ( !bSuccess && bLogSendFail ) {
		lgt = len( Str );
		MyLog( "Failed: SendText(" $ IpAddrToString( Addr ) $ ", len=" $ lgt $ " )" );
		DumpToLog( lgt, str );
	}
	return bSuccess;
}

function bool MaySend( IpAddr Addr )
{
	local int idx;
	local float Now;
	local string ip, IpPort, ban, Message;

	IpPort = IpAddrToString( Addr );
	ip = IpOnly( Addr );

	now = Level.TimeSeconds;

	// fake death on DRQA
	if ( DRQA.bActive ) {
		if ( DRQA.DragTime + TimeDeltaForDRQA >= Now ) {
			// still running
			DRQA.Count++;
			DRQA.DragTime = Now;
			return false;
		}
		// stopped
		MyLog( "NptServerQuery: DRQA attack stopped, " $ DRQA.Count $ " requests / " $ DRQA.DragTime - DRQA.FirstTime $ " seconds" );
		DRQA.bActive = false;
	}

	// RQA?
	if ( RQA.bActive && ip == RqaIP ) {
		// currently running RQA attack
		if ( RQA.DragTime + TimeDeltaForRQA >= Now ) {
			// still running
			RQA.Count++;
			RQA.DragTime = Now;
			return false;
		}
		// stopped
		MyLog( "NptServerQuery: RQA attack from " $ RqaIP $ " stopped, " $ RQA.Count $ " requests / " $ RQA.DragTime - RQA.FirstTime $ " seconds" );
		RQA.bActive = false;
		RqaIP = "";
	}

	if ( ip == CurrentIP ) {
		if ( Current.DragTime + TimeDeltaForRQA < Now ) {
			// long pause, reset counter
			Current.Count = 1;
			Current.DragTime = Now;
		}
		else {
			if ( ++Current.Count >= MaxRequestsForRQA ) {
				// this is RQA
				if ( RQA.bActive ) {
					// wow - 2 RQA attacks at once - maybe misconfiguration?
					Current.FirstTime = Current.DragTime;
					MyLog( "NptServerQuery assuming DRQA attack: " $ RQA.Count $ " requests / " $ Now - RQA.FirstTime $ " seconds from " $ RqaIP
								$ " plus " $ Current.Count $ " requests / " $ Now - Current.FirstTime $ " seconds from " $ CurrentIP );
					DRQA.bActive = true;
					DRQA.Count = RQA.Count + Current.Count;
					DRQA.FirstTime = RQA.DragTime;
					DRQA.DragTime = Now;
					return false;
				}
				RQA = Current;
				RqaIP = CurrentIP;
				RQA.bActive = true;
				RQA.FirstTime = RQA.DragTime;
				RQA.DragTime = Now;
				MyLog( "NptServerQuery RQA attack from " $ RqaIP $ ": " $ RQA.Count $ " requests / " $ RQA.DragTime - RQA.FirstTime $ " seconds" );
				if ( bBanRQA ) MyLog( "New ban entry: " $ class'NephthysProxy'.static.BanIpName( ip, "", "NptServerQuery: RQA " $ RQA.Count $ " requests / " $ RQA.DragTime - RQA.FirstTime $ " seconds", "RQA" ) );
				return false;
			}
		}
	}
	else {
		CurrentIP = ip;
		Current.Count = 1;
		Current.DragTime = Now;
	}

	// DRQA?
	if ( DRQA.DragTime + TimeDeltaForDRQA < Now ) {
		// long pause, reset counter
		DRQA.Count = 1;
		DRQA.DragTime = Now;
	}
	else if ( ++DRQA.Count >= MaxRequestsForDRQA ) {
		// this is DRQA
		DRQA.bActive = true;
		DRQA.FirstTime = DRQA.DragTime;
		DRQA.DragTime = Now;
		MyLog( "NptServerQuery DRQA attack: " $ DRQA.Count $ " requests / " $ DRQA.DragTime - DRQA.FirstTime $ " seconds" );
		return false;
	}

	// check
	Ban = class'NephthysProxy'.static.IpNameIsBanned( Message, ip, , true, true );
	if ( Len(Ban) > 0 ) {
		if ( bLogAllRequests || bLogRejected )	MyLog( "Rejected query request of " $ IpPort $ " due to ban " $ ban );
		return False;
	}

	return True;
}

//=============================================================================

function bool ParseNextQuery( string Query, out string QueryType, out string QueryValue, out string QueryRest, out string FinalPacket )
{
	local string TempQuery;
	local int ClosingSlash;

	if ( Query == "" ) return False;

	// Query should be:
	// \[type]\<value>
	if (Left(Query, 1) != "\\") return False;

	// Check to see if closed.
	ClosingSlash = InStr( Mid( Query, 1 ), "\\" );
	if ( ClosingSlash < 0 ) return False;

	TempQuery = Query;

	// Query looks like:
	// \[type]\
	QueryType = Mid( Query, 1 );
	QueryType = Left( QueryType, ClosingSlash );

	QueryRest = Mid( Query, 1 + ClosingSlash + 1 );

	if ( QueryRest == "" || Len( QueryRest ) == 1 ) QueryRest = "";
	else if ( Left( QueryRest, 1 ) == "\\" ) {
		// Query has empty value:
		// \type\\...
		;
	}
	else {
		// Query looks like:
		// \[type]\[value]		or
		// \[type]\[value]\...
		ClosingSlash = InStr( QueryRest, "\\" );
		if ( ClosingSlash >= 0 ) {
			QueryValue = Left( QueryRest, ClosingSlash );
			QueryRest = Mid( QueryRest, ClosingSlash + 1 );
		}
		else {
			QueryValue = QueryRest;
			QueryRest = "";
		}
	}
	if ( QueryRest == "" ) FinalPacket = "final";
	return True;
}

function string ParseQuery( IpAddr Addr, coerce string Query, int QueryNum, out int PacketNum )
{
	local string QueryType, QueryValue, QueryRest, ValidationString;
	local bool Result;
	local string FinalPacket;
	local int NumPlayers, AllPlayers, QueryLgt;

	QueryLgt = Len(Query);
	if ( QueryLgt > 100 ) {
		if ( bLogInvalidRequests ) MyLog( "Query length overflow attack from " $ IpAddrToString( Addr ) $ ", len=" $ QueryLgt );
		return "";
	}

	AllPlayers = GetNumPlayers();
	if ( bCountRealPlayersOnly )	NumPlayers = Level.Game.NumPlayers;
	else				NumPlayers = AllPlayers;

	Result = ParseNextQuery(Query, QueryType, QueryValue, QueryRest, FinalPacket);
	if ( !Result ) {
		if ( bLogInvalidRequests ) MyLog( "Unparseble query from " $ IpAddrToString( Addr ) $ ", len=" $ Len(Query) $ ", " $ DumpToStr(Len(Query), Query) );
		return "";
	}

	if ( QueryType == "basic" )		Result = SendQueryPacket(Addr, GetBasic(), QueryNum, ++PacketNum, FinalPacket);
	else if ( QueryType == "info" )		Result = SendQueryPacket(Addr, GetInfo2( NumPlayers ), QueryNum, ++PacketNum, FinalPacket);
	else if ( QueryType == "rules" )	Result = SendQueryPacket(Addr, GetRules(), QueryNum, ++PacketNum, FinalPacket);
	else if ( QueryType == "players" ) {
		if ( AllPlayers > 0 )		Result = SendPlayers(Addr, QueryNum, PacketNum, FinalPacket);
		else				Result = SendQueryPacket(Addr, "", QueryNum, PacketNum, FinalPacket);
	}
	else if ( QueryType == "status" ) {
		Result = SendQueryPacket(Addr, GetBasic(), QueryNum, ++PacketNum, "");
		Result = SendQueryPacket(Addr, GetInfo2( NumPlayers ), QueryNum, ++PacketNum, "");
		if ( AllPlayers == 0 ) Result = SendQueryPacket(Addr, GetRules(), QueryNum, ++PacketNum, FinalPacket);
		else {
			Result = SendQueryPacket(Addr, GetRules(), QueryNum, ++PacketNum, "");
			Result = SendPlayers(Addr, QueryNum, PacketNum, FinalPacket);
		}
	}
	else if ( QueryType == "echo" ) {
		// Respond to an echo with the same string
		Result = SendQueryPacket(Addr, "\\echo\\"$QueryValue, QueryNum, ++PacketNum, FinalPacket);
	}
	else if ( QueryType == "secure" ) {
		// protect from possible query overflow attacks
		if ( QueryLgt > 31 ) {
			if ( bLogInvalidRequests ) MyLog( "\\secure\\ query length overflow attack from " $ IpAddrToString( Addr ) $ ", len=" $ QueryLgt $ ", " $ DumpToStr(QueryLgt, Query) );
			return "";
		}
		ValidationString = "\\validate\\"$Validate(QueryValue, GameName);
		// ValidationString = "\\validate\\"$Validate(QueryValue);
		Result = SendQueryPacket(Addr, ValidationString, QueryNum, ++PacketNum, FinalPacket);
	}
	else if ( QueryType == "level_property" )	Result = SendQueryPacket(Addr, GetLevelProperty(QueryValue), QueryNum, ++PacketNum, FinalPacket);
	else if ( QueryType == "game_property" )	Result = SendQueryPacket(Addr, GetGameProperty(QueryValue), QueryNum, ++PacketNum, FinalPacket);
	else if ( QueryType == "player_property" )	Result = SendQueryPacket(Addr, GetPlayerProperty(QueryValue), QueryNum, ++PacketNum, FinalPacket);
	else if ( bLogInvalidRequests ) {
		MyLog( "Unknown query from " $ IpAddrToString( Addr ) $ ", len=" $ Len(Query) $ ", " $ DumpToStr(Len(Query), Query) );
		// don't answer anything more ...
		return "";
	}
	if ( !Result ) MyLog( "Error responding to query '" $ QueryType $ "'" );
	return QueryRest;
}

//=============================================================================

event ReceivedText( IpAddr Addr, string Text )
{
	local int lgt;

	if ( bLogAllRequests ) {
		lgt = Len( Text );
		MyLog( "ReceivedText(" $ IpAddrToString( Addr ) $ ", len=" $ lgt $ ", " $ DumpToStr(lgt, Text) $ " )" );
	}

	if ( !MaySend( Addr ) )	return; // ignore request

	Super.ReceivedText( Addr, Text );
}

event ReceivedLine( IpAddr Addr, string Line )
{
	local int lgt;

	if ( bLogAllRequests ) {
		lgt = Len( Line );
		MyLog( "ReceivedLine(" $ IpAddrToString( Addr ) $ ", len=" $ lgt $ ", " $ DumpToStr(lgt, Line) $ " )" );
	}

	// COULD EXTRACT NAME FROM Line AND LOG IT IN MAYSEND
	if ( !MaySend( Addr ) ) return; // ignore request

	Super.ReceivedLine( Addr, Line );
}

event ReceivedBinary( IpAddr Addr, int Count, byte B[255] )
{
	local string ip;

	if ( bLogAllRequests ) {
		MyLog( "Illegal ReceivedBinary(" $ IpAddrToString( Addr ) $ ", " $ DumpToStr(count, string(B)) $ " )" );
	}

	if ( !MaySend( Addr ) ) return; // ignore request

	ip = IpOnly( Addr );

	MyLog( "New ban entry: " $ class'NephthysProxy'.static.BanIpName( ip, "", "NptServerQuery: Binary", "Binary" ) );
	// ignore request
}

defaultproperties
{
				bReportPlayers=True
				bCountRealPlayersOnly=True
				bLogRejected=True
				bLogInvalidSends=True
				bLogSendFail=True
				MaxRequestsForRQA=7
				TimeDeltaForRQA=1.000000
				MaxRequestsForDRQA=20
				TimeDeltaForDRQA=1.000000
}
